@@ -83,6 +83,7 @@ precedence = infix2postfix.convert_precedence_list(cpp_precedence_list)
 class Undefined(object):
     def __int__(self): return 0
     def __str__(self): return ''
+    def __repr__(self): return '<UNDEFINED>'
 undefined = Undefined()
 
 
@@ -91,6 +92,8 @@ class Evaluator(rpn.RPN):
     An RPN calculator with infix to postfix converter, so that expressions for
     #if can be evaluated.
     """
+    re_defined_translation = re.compile(r'LOOKUP (\w+?) defined')
+
     def __init__(self):
         rpn.RPN.__init__(self)
         self.defines = {}
@@ -99,29 +102,47 @@ class Evaluator(rpn.RPN):
         self.builtins['&&'] = self.builtins['and']
         self.builtins['||'] = self.builtins['or']
 
+    def _translate_defined(self, match):
+        return u'DEFINED %s' % match.group(1)
+
     @rpn.word('LOOKUP')
     def word_LOOKUP(self, stack):
         key = self.next_word()
+        #~ print "LOOKUP", key
         if key in self.defines:
-            self.push(self.defines[key])
+            try:
+                backup = self[:]
+                value = self.eval(self.defines[key])
+                self[:] = backup # XXX better way to do this
+                self.push(value)
+            except Exception as e:
+                #~ print "RPN eval failed using directly: %r %r %s" % (key, self.defines[key], e) # XXX debug
+                self.push(self.defines[key])
         else:
             self.push(undefined)
+        #~ print "LOOKUP %r %r" % (key, self[-1])
 
     @rpn.word('DEFINED')
     def word_DEFINED(self, stack):
-        self.push(self.pop() is not undefined)  #XXX
+        key = self.next_word()
+        self.push(key in self.defines)
 
     def eval(self, expression):
         self.clear()
-        rpn_expr = infix2postfix.infix2postfix(
-                expression,
-                scanner=Scanner,
-                precedence = precedence,
-                variable_prefix='LOOKUP ')
-        #~ print >>sys.stderr, "RPN: ", rpn_expr # XXX debug
+        try:
+            rpn_expr = infix2postfix.infix2postfix(
+                    expression,
+                    scanner=Scanner,
+                    precedence = precedence,
+                    variable_prefix='LOOKUP ')
+        except ValueError, e:
+            raise PreprocessorError('error in expression: %r' % (expression,))
+        #~ print "RPN: %r" % (rpn_expr,) # XXX debug
+        # hack: replace "LOOKUP <word> DEFINED" with "DEFINED <word>"
+        rpn_expr = self.re_defined_translation.sub(self._translate_defined, rpn_expr)
         self.interpret_sequence(rpn_expr.split(' '))
         if len(self) != 1:
-            raise ValueError('error in expression: %r stack: %s' % (expression, self))
+            raise PreprocessorError('error in expression: %r stack: %s' % (expression, self))
         return self.pop()
 
 
@@ -138,7 +159,7 @@ class AnnoatatedLineWriter(object):
     def write(self, lineno, text):
         """
         Write line. It is expected to be called with lines only (ending in
-        '\n') otherwise will the markers be missplaced.
+        '\n') otherwise will the markers be misplaced.
         """
         # emit line number and file hints for the next stage
         if self.marker != lineno:
@@ -221,7 +242,7 @@ class Preprocessor(object):
         #~ print "expand -> %r" % (res)          #DEBUG
         return line.replace('##', '')
 
-    def preprocess(self, infile, outfile, filename):
+    def preprocess(self, infile, outfile, filename, include_callback=None):
         """Scan lines and process preprocessor directives"""
         self.log.info("processing %s" % filename)
         empty_lines = 0
@@ -232,8 +253,10 @@ class Preprocessor(object):
         in_comment = False
         writer = AnnoatatedLineWriter(outfile, filename)
         line = ''
+        lineno = 0
         try:
-            for n, line in enumerate(line_joiner(iter(infile).next)):
+            for line in line_joiner(iter(infile).next):
+                lineno += 1
                 #~ print "|", line.rstrip()
                 line = self.re_inlinecomment.sub('', line) #.strip()
                 if in_comment:
@@ -273,9 +296,9 @@ class Preprocessor(object):
                         my_if_was_not_hidden = False
                     continue
                 elif m.lastgroup == 'IFDEF':
-                    symbol = m.group('IFDEF_NAME')
+                    symbol = m.group('IFDEF_NAME').strip()
                     value = self.namespace.defines.has_key(symbol)
-                    self.log.debug("#ifdef %s -> %r" % (symbol, value))
+                    self.log.debug("#ifdef %r -> %r" % (symbol, value))
                     hiddenstack.append((process, my_if_was_not_hidden, if_name))
                     if_name = symbol
                     if process:
@@ -285,9 +308,9 @@ class Preprocessor(object):
                         my_if_was_not_hidden = False
                     continue
                 elif m.lastgroup == 'IFNDEF':
-                    symbol = m.group('IFNDEF_NAME')
+                    symbol = m.group('IFNDEF_NAME').strip()
                     value = not self.namespace.defines.has_key(symbol)
-                    self.log.debug("#ifndef %s -> %r" % (symbol, value))
+                    self.log.debug("#ifndef %r -> %r" % (symbol, value))
                     hiddenstack.append((process, my_if_was_not_hidden, if_name))
                     if_name = symbol
                     if process:
@@ -297,27 +320,29 @@ class Preprocessor(object):
                         my_if_was_not_hidden = False
                     continue
                 elif m.lastgroup == 'ELSE':
-                    self.log.debug("#else %s" % (if_name))
+                    self.log.debug("#else %r" % (if_name,))
                     if my_if_was_not_hidden:
                         process = not process
                     continue
                 elif m.lastgroup == 'ENDIF':
-                    self.log.debug("#endif %s" % (if_name))
+                    self.log.debug("#endif %r" % (if_name,))
                     (process, my_if_was_not_hidden, if_name) = hiddenstack.pop()
                     continue
                 elif not process:
                     continue
                 elif m.lastgroup == 'INCLUDE':
                     include_name = m.group('INC_NAME')
-                    self.log.debug('including "%s"' % include_name)
+                    self.log.debug('including "%s"' % (include_name,))
                     for location in self.include_path:
                         path = os.path.normpath(os.path.join(location, include_name))
                         if os.path.exists(path):
-                            self.preprocess(codecs.open(path, 'r', 'utf-8'), outfile, path)
+                            if include_callback is not None:
+                                include_callback(path)
+                            self.preprocess(codecs.open(path, 'r', 'utf-8'), outfile, path, include_callback)
                             writer.marker = None  # force marker output
                             break
                     else:
-                        raise PreprocessorError('include file %r not found' % (include_name))
+                        raise PreprocessorError('include file %r not found' % (include_name,))
                     continue
                 elif m.lastgroup == 'MACRO':
                     name = m.group('MACRO_NAME')
@@ -327,7 +352,7 @@ class Preprocessor(object):
                     else:
                         definition = ''
                     if self.macros.has_key(name):
-                        self.log.warn("%r redefinition ignored" % name)
+                        self.log.warn("%r redefinition ignored" % (name),)
                     else:
                         # prepare the macro value to be used as format string
                         # (python's % operator)
@@ -351,14 +376,14 @@ class Preprocessor(object):
                     else:
                         definition = ''
                     if self.namespace.defines.has_key(symbol):
-                        self.log.warn("%r redefinition ignored" % name)
+                        self.log.warn("%r redefinition ignored" % (symbol,))
                     else:
                         self.namespace.defines[symbol] = definition
                         self.log.debug("defined %r => %r" % (symbol, self.namespace.defines[symbol]))
                     continue
                 elif m.lastgroup == 'UNDEF':
-                    symbol = m.group('UNDEF_NAME')
-                    self.log.debug("undefined %s" % symbol)
+                    symbol = m.group('UNDEF_NAME').strip()
+                    self.log.debug("undefined %s" % (symbol,))
                     del self.namespace.defines[symbol]
                     continue
                 elif m.lastgroup == 'NONPREPROC':
@@ -372,12 +397,24 @@ class Preprocessor(object):
                         continue
                 else:
                     empty_lines = 0
-                writer.write(n+1, line)
+                writer.write(lineno, line)
+            # at the end of the loop, check if there were unbalanced #if's
+            if hiddenstack:
+                raise PreprocessorError('missing #endif')
+
+        except PreprocessorError, e:
+            # annotate exception with location in source file
+            e.line = lineno
+            e.filename = filename
+            e.text = line
+            self.log.info('error while processing "%s"' % (line.strip(),))
+            raise
         except:
-            self.log.info('error while processing "%s"' % line.strip())
+            self.log.info('error while processing "%s"' % (line.strip(),))
             raise
         else:
-            self.log.info('done "%s"' % filename)
+            self.log.info('done "%s"' % (filename),)
+
 
 class Discard(object):
     """File like target object that consumes and discards all data"""
@@ -385,7 +422,7 @@ class Discard(object):
         """dummy write"""
 
 
-if __name__ == '__main__':
+def main():
     import sys
     from optparse import OptionParser
     logging.basicConfig()
@@ -421,6 +458,11 @@ if __name__ == '__main__':
                       metavar = "PATH",
                       default = [],
                       help="Add directory to the search path list for includes")
+    parser.add_option("--dependency-scan",
+                      action = "store_true",
+                      dest = "dependency_scan_only",
+                      default = False,
+                      help = "just print names of includes, not actual content")
 
     (options, args) = parser.parse_args()
 
@@ -445,8 +487,12 @@ if __name__ == '__main__':
         infilename = '<stdin>'
         infile = codecs.getreader("utf-8")(sys.stdin)
     else:
-        infilename = args[0]
-        infile = codecs.open(infilename, 'r', 'utf-8')
+        try:
+            infilename = args[0]
+            infile = codecs.open(infilename, 'r', 'utf-8')
+        except IOError, e:
+            sys.stderr.write('cpp: %s: File not found\n' % (infilename,))
+            sys.exit(1)
 
     cpp = Preprocessor()
     # extend include search path
@@ -458,7 +504,29 @@ if __name__ == '__main__':
         else:
             symbol, value = definition, '1'
         cpp.namespace.defines[symbol] = value
+
     # process files now
     if options.preload:
         cpp.preprocess(open(options.preload), Discard(), options.preload)
-    cpp.preprocess(infile, outfile, infilename)
+
+    if options.dependency_scan_only:
+        # add a callback that writes out filenames of includes
+        # remember outfile as we're changing it below
+        def print_include(path, outfile=outfile):
+            outfile.write('%s\n' % (path,))
+        outfile = Discard() # discard following output
+    else:
+        print_include = None
+
+    try:
+        cpp.preprocess(infile, outfile, infilename, print_include)
+    except PreprocessorError, e:
+        sys.stderr.write('%s:%s: %s\n' % (e.filename, e.line, e))
+        if options.debug:
+            if hasattr(e, 'text'):
+                sys.stderr.write('%s:%s: input line: %r\n' % (e.filename, e.line, e.text))
+        sys.exit(1)
+
+
+if __name__ == '__main__':
+    main()
